@@ -25,11 +25,11 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <getopt.h>
-#include <assert.h>
 
 #include <libtasn1.h>
 
-#define program_name "asn1Parser"
+#include <progname.h>
+#include <version-etc.h>
 
 /* This feature is available in gcc versions 2.5 and later.  */
 #if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 5)
@@ -58,7 +58,7 @@ Mandatory arguments to long options are mandatory for short options too.\n\
   -n, --name=NAME       array name\n\
   -h, --help            display this help and exit\n\
   -v, --version         output version information and exit\n");
-     printf ("Report bugs to "PACKAGE_BUGREPORT);
+      emit_bug_reporting_address ();
     }
   exit (status);
 }
@@ -84,6 +84,8 @@ main (int argc, char *argv[])
   char errorDescription[ASN1_MAX_ERROR_DESCRIPTION_SIZE];
   int parse_result = ASN1_SUCCESS;
 
+  set_program_name (argv[0]);
+
   opterr = 0;			/* disable error messages from getopt */
 
   while (1)
@@ -109,9 +111,8 @@ main (int argc, char *argv[])
 	  usage (EXIT_SUCCESS);
 	  break;
 	case 'v':		/* VERSION */
-	  printf(program_name" "PACKAGE" " VERSION"\n");
-	  printf("Copyright (C) 2017-2019 Free Software Foundation, Inc.\n\n");
-	  printf("Written by Fabio Fiorina\n");
+	  version_etc (stdout, program_name, PACKAGE, VERSION,
+		       "Fabio Fiorina", NULL);
 	  free (outputFileName);
 	  free (vectorName);
 	  exit (0);
@@ -120,22 +121,12 @@ main (int argc, char *argv[])
 	  checkSyntaxOnly = 1;
 	  break;
 	case 'o':		/* OUTPUT */
-	  assert(optarg != NULL);
-	  outputFileName = strdup(optarg);
-	  if (outputFileName == NULL)
-	    {
-	      fprintf(stderr, "Memory error\n");
-	      exit(1);
-            }
+	  outputFileName = (char *) malloc (strlen (optarg) + 1);
+	  strcpy (outputFileName, optarg);
 	  break;
 	case 'n':		/* VECTOR NAME */
-	  assert(optarg != NULL);
-	  vectorName = strdup(optarg);
-	  if (vectorName == NULL)
-	    {
-	      fprintf(stderr, "Memory error\n");
-	      exit(1);
-            }
+	  vectorName = (char *) malloc (strlen (optarg) + 1);
+	  strcpy (vectorName, optarg);
 	  break;
 	case '?':		/* UNKNOW OPTION */
 	  fprintf (stderr,
