@@ -4,7 +4,7 @@
 
 #include <libtasn1.h>
 
-const asn1_static_node_t Test_tree_asn1_tab[] = {
+const asn1_static_node Test_tree_asn1_tab[] = {
   { "TEST_TREE", 536875024, NULL },
   { NULL, 1610612748, NULL },
   { "iso", 1073741825, "1"},
@@ -87,10 +87,10 @@ const asn1_static_node_t Test_tree_asn1_tab[] = {
   { NULL, 2056, "2"},
   { "int2", 1610620931, NULL },
   { NULL, 4104, "3"},
-  { "str1", 1610620930, "PrintableString"},
+  { "str1", 1610620959, NULL },
   { NULL, 4104, "1"},
-  { "str2", 1073741826, "UniversalString"},
-  { "str3", 536879106, "UniversalString"},
+  { "str2", 1073741856, NULL },
+  { "str3", 536879136, NULL },
   { NULL, 2056, "2"},
   { "Sequence1", 1610612741, NULL },
   { "int1", 1610915843, NULL },
@@ -155,11 +155,11 @@ const asn1_static_node_t Test_tree_asn1_tab[] = {
   { "GeneralString2", 1610620955, NULL },
   { NULL, 2056, "2"},
   { "X520LocalityName", 1610612754, NULL },
-  { "teletexString", 1073741826, "TeletexString"},
-  { "printableString", 1073741826, "PrintableString"},
-  { "universalString", 1073741826, "UniversalString"},
-  { "utf8String", 1073741826, "UTF8String"},
-  { "bmpString", 2, "BMPString"},
+  { "teletexString", 1073741854, NULL },
+  { "printableString", 1073741855, NULL },
+  { "universalString", 1073741856, NULL },
+  { "utf8String", 1073741858, NULL },
+  { "bmpString", 33, NULL },
   { "id-Test", 805306380, NULL },
   { NULL, 1073741825, "1"},
   { NULL, 1073741825, "2"},
